from gurobipy import *
from TempGraph import *
from temp_iso_cliques import *





def max_degree(vertex_max_cliques):
    # type: (list[collections.OrderedDict[int, Vertex]]) -> int

    max_size = 0

    for clique in vertex_max_cliques:
        max_size = max(max_size, len(clique))

    return max_size



def compare_exessive(temp_graph, c):
    # type: (Tempgraph, float) -> bool

    correct = True

    for a in range(temp_graph.t_min, temp_graph.t_max + 1):
        static_graph = temp_graph.getStaticGraph(a, a)
        static_graph.sort_by_degree()
        for b in range(a, temp_graph.t_max):
            if a != b:
                static_graph.add_another_timestamp(temp_graph, b)
                if static_graph.numEdges == 0:
                    break
                static_graph.sort_by_degree()

            vertex_max_cliques = run_stages(temp_graph, static_graph, c)

            our_max = max_degree(vertex_max_cliques)

            ilp_max = instance_to_ilp(temp_graph, a, b, c, temp_graph.mode, "tmp.lp")

            exists = False
            for clique in vertex_max_cliques:
                if set([vertex.name for vertex in clique.values()]) == set(ilp_max):
                    exists = True
                    break

            if not exists:
                print("Different on [" + str(a) + "," + str(b) + "] with c: " + str(c))
                print("Our: " + str(our_max) + " ILP: " + str(ilp_max))
                correct = False

            # TODO: write lp formulation to file
            #       read it and add single contrain
            #       break if size = 0

            # TODO:
            #  constr:
            #  all vertices in ilp_max added together shall be smaller than their size [not same clique]
            #  also no suberset, since this would mean bigger clique
            #  this way we can do it exhaustively

            #if our_max != len(ilp_max):
            #    correct = False
            #    print("Different on [" + str(a) + "," + str(b) + "] with c: " + str(c))
            #    print("Our: " + str(our_max) + " ILP: " + str(ilp_max))
    return correct


def compare_ilp(temp_graph, c, t_min=-1, t_max=-1):
    # type: (Tempgraph, float, int, int) -> bool


    output = True
    correct = True

    if t_min == -1:
        t_min = temp_graph.t_min

    if t_max == -1:
        t_max = temp_graph.t_max



    for a in range(t_min, t_max + 1):
        static_graph = temp_graph.getStaticGraph(a, a)
        static_graph.sort_by_degree()
        for b in range(a, a + 10):
        #for b in range(a, t_max):

            if a != b:
                static_graph.add_another_timestamp(temp_graph, b)
                if static_graph.numEdges == 0:
                    break
                static_graph.sort_by_degree()
            if output:
                print("compare for intervall: [" + str(a) + "," + str(b) + "] c: " + str(c) + " mode: " + str(temp_graph.mode))

            start = time.time()

            vertex_max_cliques = run_stages(temp_graph, static_graph, c)

            our_max = max_degree(vertex_max_cliques)

            ilp_max = instance_to_ilp(temp_graph, a, b, c, temp_graph.mode)

            if output:
                end = time.time()
                print("Compare intervall in %.4f" % (end - start))

            # Give out error if not of same size and ilp_max is bigger than one, otherwise the ILP just selects an
            # arbitrary isolated vertex if possible. Our program may not do this, since we may break out before
            # even without the reduction rule of not eliminationg size 1 cliques.
            if our_max != len(ilp_max) and len(ilp_max) > 1:
                correct = False
                print("Different on [" + str(a) + "," + str(b) + "] with c: " + str(c) + " with mode:" + str(temp_graph.mode))
                print("Our: " + str(our_max) + " ILP: " + str(ilp_max))


    return correct


def instance_to_ilp(temporal_graph, a, b, c, mode, extrafile=""):
    # type: (TempGraph, int, int, float, int) -> list[int]

    m = Model()
    variables = collections.OrderedDict()

    # init variables for vertices
    e1 = LinExpr()

    for i in temporal_graph.vertexList.values():
        if str(i.name) not in variables.keys():
            xi = m.addVar(vtype=GRB.BINARY, name=("x_" + str(i.name)))
            variables[str(i.name)] = xi
        for j in temporal_graph.vertexList.values():
            if i.name > j.name:
                if str(j.name) not in variables.keys():
                    xj = m.addVar(vtype=GRB.BINARY, name=("x_" + str(i.name)))
                    variables[str(j.name)] = xj
                xij = m.addVar(vtype=GRB.BINARY, name="X_" + str(i.name) + "_" + str(j.name))
                variables[str(i.name) + "_" + str(j.name)] = xij
                xi = variables[str(i.name)]
                xj = variables[str(j.name)]
                m.addConstr(xij <= xi)
                m.addConstr(xij <= xj)
                m.addConstr(xi + xj - 1 <= xij)
                curr_timestamp = a
                while curr_timestamp <= b:          # x_i * x_j < e_ijt
                    if temporal_graph.edgeList[i].keys().__contains__(curr_timestamp):
                        if not temporal_graph.edgeList[i][curr_timestamp].__contains__(j):
                            m.addConstr(xij <= 0)
                    else:
                        m.addConstr(xij <= 0)
                        m.addConstr(xi <= 0)
                    curr_timestamp += 1

        xi = variables[str(i.name)]
        e1 += xi

    # Maximize
    m.setObjective(e1, GRB.MAXIMIZE)

    # mode constrains begin here:

    ilp_constrains(temporal_graph, a, b, c, variables, mode, m)

    if extrafile != "":
        m.write(extrafile)

    m.setParam('OutputFlag', False)

    m.optimize()

    opt = []        # type: list[int]

    try:
        for i in temporal_graph.vertexList.values():
            xi = variables[str(i.name)]
            # AttributeError: b"Unable to retrieve attribute 'X'" in line below
            # xi not initialized??
            if xi.X > 0.5:      # if it is 1
                opt.append(i.name)
    except AttributeError:
        # We have an infeasible model.
        return []

    return opt


def ilp_constrains(temporal_graph, a, b, c, variables, mode, m):

    if mode == 0:
        constrains_alltime_max(temporal_graph, a, b, c, variables, m)
    if mode == 1:
        constrains_max_usually(temporal_graph, a, b, c, variables, m)
    if mode == 2:
        constrains_usally_average(temporal_graph, a, b, c, variables, m)
    if mode == 3:
        constrains_alltime_average(temporal_graph, a, b, c, variables, m)
    if mode == 4:
        constrains_average_alltime(temporal_graph, a, b, c, variables, m)
    if mode == 5:
        constrains_usally_max(temporal_graph, a, b, c, variables, m)


def constrains_max_usually(temporal_graph, a, b, c, variables, m):

    for i in temporal_graph.vertexList.values():
        xi = variables[str(i.name)]
        e1 = LinExpr()
        curr_timestamp = a
        while curr_timestamp <= b:
            e1 += xi * temporal_graph.getDegree_for_timestamp(i, curr_timestamp)
            for j in temporal_graph.vertexList.values():
                if i.name > j.name:
                    xij = variables[str(i.name) + "_" + str(j.name)]
                    e1 += -xij
                elif i.name < j.name:
                    xij = variables[str(j.name) + "_" + str(i.name)]
                    e1 += -xij
            curr_timestamp += 1
        # TODO: here is smaller, but we are only allowed to use smaller equal.
        # TODO: replace with correct term as soon as found
        m.addConstr(e1 <= c*(b + 1 - a))



def constrains_usally_average(temporal_graph, a, b, c, variables, m):

    # we could also just add the weights of all vertices and substract the xi equal 1 (b + 1 -a) times [prolly done intern]

    e1 = LinExpr()
    e2 = LinExpr()
    curr_timestamp = a
    while curr_timestamp <= b:
        for i in temporal_graph.vertexList.values():
            xi = variables[str(i.name)]
            if curr_timestamp == a:
                e2 += xi * c * (b + 1 - a)
            e1 += xi * temporal_graph.getDegree_for_timestamp(i, curr_timestamp)
            for j in temporal_graph.vertexList.values():
                if i.name > j.name:
                    xij = variables[str(i.name) + "_" + str(j.name)]
                    e1 += -xij
                elif i.name < j.name:
                    xij = variables[str(j.name) + "_" + str(i.name)]
                    e1 += -xij

        curr_timestamp += 1
    # TODO: here as well. we are not allowed to use "smaller than", but it is necessary for our case.
    m.addConstr(e1 <= e2)

def constrains_alltime_average(temporal_graph, a, b, c, variables, m):
    curr_timestamp = a
    while curr_timestamp <= b:
        e1 = LinExpr()
        e2 = LinExpr()
        for i in temporal_graph.vertexList.values():
            degree = temporal_graph.getDegree_for_timestamp(i, curr_timestamp)

            if degree > 0:
                xi = variables[str(i.name)]
                e1 += xi * degree
                for j in temporal_graph.vertexList.values():
                    if i.name > j.name:
                        xij = variables[str(i.name) + "_" + str(j.name)]
                        e1 += -xij
                    elif i.name < j.name:
                        xij = variables[str(j.name) + "_" + str(i.name)]
                        e1 += -xij
                e2 += xi * c
        m.addConstr(e1 + 1 <= e2)
        curr_timestamp += 1



def  constrains_average_alltime(temporal_graph, a, b, c, variables, m):


    e2 = LinExpr()
    e3 = LinExpr()
    for i in temporal_graph.vertexList.values():
        xi = variables[str(i.name)]
        e3 += xi * c
        max_degree_i = m.addVar(vtype=GRB.INTEGER, name=("max_degree_x_" + str(i.name)))
        e2 += max_degree_i
        variables["max_degree_x" + str(i.name)] = max_degree_i
        curr_timestamp = a
        while curr_timestamp <= b:
            e1 = LinExpr()
            e1 += xi * temporal_graph.getDegree_for_timestamp(i, curr_timestamp)
            for j in temporal_graph.vertexList.values():
                if i.name > j.name:
                    xij = variables[str(i.name) + "_" + str(j.name)]
                    e1 += -xij
                elif i.name < j.name:
                    xij = variables[str(j.name) + "_" + str(i.name)]
                    e1 += -xij
            m.addConstr(e1 <= max_degree_i)
            curr_timestamp += 1
    m.addConstr(e2 <= e3)
    # TODO: also got the problem with "smaller equal" instead of smaller


def constrains_usally_max(temporal_graph, a, b, c, variables, m):

    e2 = LinExpr()
    curr_timestamp = a
    while curr_timestamp <= b:
        max_degree_timestamp_a = m.addVar(vtype=GRB.INTEGER, name=("max_degree_timestamp_" + str(a)))
        e2 += max_degree_timestamp_a
        for i in temporal_graph.vertexList.values():
            xi = variables[str(i.name)]
            e1 = LinExpr()
            e1 += xi * temporal_graph.getDegree_for_timestamp(i, curr_timestamp)
            for j in temporal_graph.vertexList.values():
                if i.name > j.name:
                    xij = variables[str(i.name) + "_" + str(j.name)]
                    e1 += -xij
                elif i.name < j.name:
                    xij = variables[str(j.name) + "_" + str(i.name)]
                    e1 += -xij
            m.addConstr(e1 <= max_degree_timestamp_a)
        curr_timestamp += 1
    m.addConstr(e2 <= c*(b + 1 - a))
    # TODO: also got the problem with "smaller equal" instead of smaller
    return


def constrains_alltime_max(temporal_graph, a, b, c, variables, m):
    # alltime_max:
    curr_timestamp = a
    while curr_timestamp <= b:
        for i in temporal_graph.vertexList.values():
            xi = variables[str(i.name)]
            e1 = LinExpr()
            e1 += xi * temporal_graph.getDegree_for_timestamp(i, curr_timestamp)
            for j in temporal_graph.vertexList.values():
                if i.name > j.name:
                    xij = variables[str(i.name) + "_" + str(j.name)]
                    e1 += -xij
                elif i.name < j.name:
                    xij = variables[str(j.name) + "_" + str(i.name)]
                    e1 += -xij
            m.addConstr(e1 <= c)

        curr_timestamp += 1






























