import argparse
import sys
import time
import trimming
import enumeration
from StaticGraph import StaticGraph
from TempGraph import TempGraph


def main():
    #   Example Code:
    parser = argparse.ArgumentParser(usage="Usage: cat <temporal graph> | ./temporal-cliques.py [options]")
    # parser.add_argument("-k", "--k", type=int, default=1, help="k for kplexes, default is 1.")
    parser.add_argument("-d", "--delta", type=int, default=1, help="Delta, default is 1.")
    parser.add_argument("-e", "--extended-time", action="store_true", default=False,
                        help="If selected, kplexes will be computed on interval [t_min-delta, t_max+delta] for consistency with temporal-cliques.")
    parser.add_argument("-c", "--c",  default=5, help="Input Parameter c, default is 5.")
    # parser.add_argument("-m", "--min-kplex-size", type=int, default=1, help="Minimum size for enumerated kplexes.")
    # parser.add_argument("-o", "--kplex-file", help="Path of file to write enumerated kplexes to.")
    # parser.add_argument("-s", "--statistics-file", help="Path of file to write computation statistics to.")
    parser.add_argument("-n", "--filename", default="", help="file name of data.")
    parser.add_argument("-algo", "--algo", default=0, help="index of algorithm used.")
    args = parser.parse_args()

    # Read temporal graph from stdin
    # Also use delta reduction
    g = read_graph(args.filename)




    # test cover here

    k = 4

    covers = enumeration.vertex_cover(g, k)

    enumeration.print_vertex_covers(covers)

    # TODO: create static/temp graph and check correctness of isolated subset





def read_graph(filename):

    start = time.time()

    g = StaticGraph("G")

    with open(filename) as f:
        content = f.readlines()
    content = [line.rstrip('\n') for line in content]


    contents = content[0].split()
    u, v = int(contents[0]), int(contents[1])


    g.addEdge(u, v)


    for line in content:
        contents = line.split()

        if len(contents) == 2:

            u, v = int(contents[0]), int(contents[1])

            if u == v:
                print >> sys.stderr, "Ignoring self-loop on vertex %s at time %s" % (u, v)
                continue

            g.addEdge(u, v)


    end = time.time()
    print("Read input %.4f" % (end - start))
    return g



if __name__ == "__main__":
    sys.exit(main())
