import datetime
import time

from TempGraph import *
from temp_iso_cliques import read_temporal_graph, algorithm, map_mode


def map_delta(d):
    if d == 0:
        return "0"
    if d == 1:
        return "5^3"
    if d == 2:
        return "5^5"
    if d == 3:
        return "5^7"

def algorithm_clipped(filename, timestamp, c, mode, delta, outputname, output, period=20, version=1, d_2 = 0):

    start = time.time()

    del_small_clique = 1
    temporal_graph_iter = read_temporal_graph(filename, delta, period, mode, del_small_clique, version)
    num_cand_set = 0
    num_clique_set = 0
    list_time_max_cliques = []
    num_timestamps = 0
    num_iter_timestamps = 0

    num_graph = 0
    num_vertices = 0
    num_edges = 0
    for graph in temporal_graph_iter:

        if output and graph.t_max - graph.t_min > 300:
            print("new_graph_" + str(num_graph) + ": " + str(graph.t_min) + " " + str(graph.t_max))
        num_vertices = graph.numVertices
        num_graph += 1
        num_edges += graph.numEdges
        num_timestamps += graph.t_max - graph.t_min + 1
        time_max_cliques = algorithm(graph, c, timestamp, output)
        num_iter_timestamps += graph.num_timestamps_iter
        list_time_max_cliques.append(time_max_cliques)
        num_cand_set += graph.num_candidate_sets
        num_clique_set += graph.num_clique_before_time_max

    end = time.time()

    overall_size = 0
    num_time_max = 0
    max_size = 0
    max_lifetime = 0
    max_lifetime_array = [0, 0, 0, 0, 0, 0, 0, 0]
    overall_lifetime = 0

    for time_max_cliques in list_time_max_cliques:
        for a in time_max_cliques.keys():
            for b in time_max_cliques[a].keys():
                lifetime = b + 1 - a
                max_lifetime = max(max_lifetime, lifetime)
                for clique in time_max_cliques[a][b]:
                    len_clique = len(clique)
                    num_time_max += 1
                    max_size = max(max_size, len_clique)
                    overall_size += len_clique
                    overall_lifetime += lifetime
                    if lifetime < len(max_lifetime_array):
                        max_lifetime_array[lifetime] += 1


    avg_size = 0
    average_lifetime = 0
    time_timestamp = (end - start)
    time_clique = (end - start)
    if num_time_max != 0:
        avg_size = overall_size/num_time_max
        average_lifetime = overall_lifetime/num_time_max
        time_timestamp = (end - start) / num_iter_timestamps
        time_clique = (end - start) / num_time_max


    print("Filename: " + str(filename) + " n: " + str(num_vertices) + " M: " + str(num_edges) + " C: " + str(c) + " Mode: " + str(map_mode(mode)) + " delta: " + str(delta) + " delta_2: " + str(map_delta(d_2)) +
          " timestamps: " + str(num_timestamps) +
          "\n" + "Time: " + str(end - start) +
          " time/clique: " + str(time_timestamp) +
          " time/timestamp: " + str(time_clique) +
          " iterated timestamps: " + str(num_iter_timestamps) + "\n" +
          "num_time_max: " + str(num_time_max) +
          " max_size: " + str(max_size) +
          " avg_size: " + str(avg_size) +
          " num_cand_set: " + str(num_cand_set) +
          " num_clique_set: " + str(num_clique_set) + "\n" +
          "max_lifetime   : " + str(max_lifetime) +
          " average_lifetime: " + str(average_lifetime) +
          " max_lifetime[1]: " + str(max_lifetime_array[1]) +
          " max_lifetime[2]: " + str(max_lifetime_array[2]) +
          " max_lifetime[3]: " + str(max_lifetime_array[3]) +
          " max_lifetime[4]: " + str(max_lifetime_array[4]) +
          "\n")

    with open(outputname, "a") as f:
        f.write(str(filename) + ", " +
                str(num_vertices) + ", " +
                str(num_edges) + ", " +
                str(end - start) + ", " +
                str(time_clique) + ", " +
                str(num_iter_timestamps) + ", " +
                str(time_timestamp) + ", " +
                str(c) + ", " +
                str(delta) + "," +
                str(map_delta(d_2)) + "," +
                map_mode(mode) + "," +
                str(num_time_max) + "," +
                str(max_size) + "," +
                str(avg_size) + "," +
                str(num_cand_set) + "," +
                str(num_clique_set) + "," +
                str(max_lifetime_array[1]) + "," +
                str(max_lifetime_array[2]) + "," +
                str(max_lifetime_array[3]) + "," +
                str(max_lifetime_array[4]) + "," +
                str(max_lifetime) + "," +
                str(average_lifetime) + "," +
                str(num_timestamps) + "," +
                "\n")

    return list_time_max_cliques

def report_failure(file, path, max_time, c, mode, d, t, d_2):
    file.write(str(path) + ",0,0," + str(max_time) + ",0,0,0," + str(c) + "," + str(d) + "," + str
        (map_delta(d_2)) + "," + str(map_mode(mode)) +
                                                                                                   ",0,0,0,0,0,0,0,0,0,0,0," + str(t) + "\n")


def algorithm_clipped_wrapped(*args, **kwargs):
    try:
        return algorithm_clipped(*args, **kwargs)
    except Exception as e:
        with open("UNEXPECTED_ERRORS.txt", mode="a") as f:
            f.write(str(datetime.datetime.now()) + ":\n" + str(args) + str(kwargs) + str(e) + "\n\n")


def run_on_server(path, period=20):

    version = 1
    t = -1
    max_time = 3600
    c_values = [0.001, 1, 5, 25 ,125]
    #modes = [0 ,1 ,2 ,3 ,4]
    modes = [2]

    # files = os.listdir(folder)
    now = datetime.datetime.now()
    outputname = "comparison_" + str(now.strftime("%Y-%m-%d")) + ".csv"
    with open(outputname, "a") as f:
        f.write \
            ("filename, n, m, runtime, time per clique, num timestamps iterated, time per timestamp, c, delta, delta factor, mode, num_cliques, largest_clique, avg_size, candidate_sets," +
                "overall_cliques, lifetime_1, lifetime_2, lifetime_3, lifetime_4, max_lifetime, average_lifetime, timestamp\n")

        def run(mode):

            output = False
            temporal_graph_iter = read_temporal_graph(path, 0, period, 0, 1, version)
            lifetime = 0
            time_edges = 0
            for graph in temporal_graph_iter:
                lifetime += graph.t_max + 1 - graph.t_min
                time_edges += graph.numEdges
            temporal_graph_iter = None              # free memory

            delta_values = [0]
            for i in range(3, 7+ 1, 2):
                delta_values.append(int(pow(5, i) * (lifetime * period / (5 * time_edges * period))))

            d_2 = 0

            fails = [0] * len(c_values)

            for d in delta_values:
                index_c = 0
                for c in c_values:

                    if fails[index_c] == 1:
                        continue  # skip remaining delta values for same c value

                    p = multiprocessing.Process(target=algorithm_clipped_wrapped,
                                                args=[path, t, c, mode, d, outputname, output, period, version, d_2])
                    p.start()

                    # Wait for 1/2 hour or until process finishes
                    p.join(max_time)

                    # If thread is still active
                    if p.is_alive():
                        report_failure(f, path, max_time, c, mode, d, t, d_2)
                        fails[index_c] = 1
                        index_c += 1
                        # Terminate
                        p.terminate()
                        p.join()
                        # Removed break since we don't want to skip for same delta and higher c anymore
                        #break  # skip remaining c values
                    index_c += 1
                d_2 += 1


        workers = []
        for mode in modes:
            worker = threading.Thread(
                target=run,
                args=(mode,)
            )
            worker.start()
            workers.append(worker)

        for worker in workers:
            worker.join()


def main():
    period = 1
    path = "data/dnc.txt"
    run_on_server(path, period)
    period = 20
    path = "data/highschool-2011.txt"
    run_on_server(path, period)
    path = "data/highschool-2012.txt"
    run_on_server(path, period)
    path = "data/highschool-2013.txt"
    run_on_server(path, period)
    path = "data/infectious.txt"
    run_on_server(path, period)
    path = "data/tij_pres_LH10.dat"
    run_on_server(path, period)


if __name__ == "__main__":
    sys.exit(main())

