import argparse
import sys
import time
from itertools import starmap
import cProfile
import datetime
import threading

import pytest


import multiprocessing
import trimming
import enumeration
from StaticGraph import *
from TempGraph import *
from solution import *
from utils import *
import os
from subprocess import PIPE, Popen
import collections
from BromKerbosch import *
from server_tests import *


def first(iterable):
    return next(iter(iterable))


# This function is for testing the correctness of creating delta graphes manually!
def create_delta_graph(filename, delta):
    # type: (str, int) -> None

    list_g = read_temporal_graph(filename, delta)

    first(list_g).print_graph()



def run_interval(filename, delta, c, a, b):

    period = 1          # should only create one graph for period 1

    temporal_graph = first(read_temporal_graph(filename, delta, period))

    static_graph = temporal_graph.getStaticGraph(a, b)

    static_graph.sort_by_degree()

    result = run_stages(temporal_graph, static_graph, c)

    return result



def test_example1():
    score = 0
    correct1 = [[0, 1, 2], [0, 3], [3, 4, 5]]
    correct2 = [[0, 1, 2], [3, 4, 5]]
    filename = "own_files/clique2.txt"
    delta = 2
    c = 2.01
    a = 1
    b = 1

    cliques = vertex_maximal_cliques(run_interval(filename, delta, c, a, b))
    clique1 = []
    for clique in cliques:
        cur_clique = []
        for vertex in clique.values():
            cur_clique.append(int(vertex.name))
        clique1.append(sorted(cur_clique))
    clique1 = sorted(clique1)

    c = 2
    cliques = vertex_maximal_cliques(run_interval(filename, delta, c, a, b))
    clique2 = []
    for clique in cliques:
        cur_clique = []
        for vertex in clique.values():
            cur_clique.append(int(vertex.name))
        clique2.append(sorted(cur_clique))
    clique2 = sorted(clique2)
    assert clique1 == correct1
    assert clique2 == correct2


@pytest.mark.parametrize("c", [2, 3])
def test_example2(c):

    filename = "own_files/clique2.txt"
    delta = 2

    a = 1
    b = 1

    period = 1

    temporal_graph = first(read_temporal_graph(filename, delta, period))

    static_graph = temporal_graph.getStaticGraph(a, b)

    assert compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, c, True)


def test_example3():


    filename = "own_files/random_n_17_m_15_noise_8.txt"
    delta = 1

    period = 1

    temporal_graph = first(read_temporal_graph(filename, delta, period))

    static_graph = temporal_graph.getStaticGraph(0, 0)
    assert compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, 3, True)


def test_example4():

    filename = "own_files/random_n_10_m_25_noise_4.txt"
    delta = 1

    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    static_graph = temporal_graph.getStaticGraph(0, 0)
    assert compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, 5, True)



def test_example5():

    filename = "own_files/random_n_10_m_30_noise_0.txt"
    delta = 1

    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    static_graph = temporal_graph.getStaticGraph(0, 0)
    assert compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, 2, True)



def test_example6():


    filename = "own_files/random_n_10_m_40_noise_4.txt"
    delta = 1

    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    static_graph = temporal_graph.getStaticGraph(0, 0)
    assert compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, 2, True)


def test_example8():
    filename = "own_files/random_n_17_m_15_noise_8.txt"
    delta = 1

    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    static_graph = temporal_graph.getStaticGraph(0, 0)
    assert compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, 3, True)


def test_example9():
    filename = "own_files/graph_split.txt"
    delta = 1

    period = 1
    iter_g = read_temporal_graph(filename, delta, period)

    g0 = next(iter_g)
    g1 = next(iter_g)

    g0.print_graph()
    g1.print_graph()

    assert g0.has_edge(0, 0, 1)
    assert g0.has_edge(0, 0, 3)
    assert g0.has_edge(0, 1, 2)
    assert g1.has_edge(9, 0, 1)
    assert g1.has_edge(10, 2, 3)
    assert g1.has_edge(10, 0, 1)


def solution_has_clique(solution, clique, a, b):
    # type: (collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]], set, int, int) -> bool

    # we check whether given instance of solution contains the given clique for intervall exactly [a,b]!!

    if solution.keys().__contains__(a):
        if solution[a].keys().__contains__(b):
            for time_max_clique in solution[a][b]:
                if set([vertex.name for vertex in time_max_clique.values()]) == clique:
                    return True

    return False



def assert_temporal_cliques_equal(expected, result):
    try:
        assert len(expected) == num_time_max_cliques(result), "Result has wrong number of cliques"

        for entry in expected:
            assert solution_has_clique(result, entry.clique, entry.a, entry.b), "Result is missing clique {}".format(entry)
    except AssertionError:
        print("Result:")
        print_time_max_cliques(result)
        raise


def test_alltime_avg():
    filename = "own_files/alltime_avg.txt"
    delta = 0
    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    score = 0

    temporal_graph.mode = 3  # alltime_avg
    temporal_graph.del_small_clique = 1

    c = 2.01

    timestamps = -1
    output = False
    time_max_cliques = algorithm(temporal_graph, c, timestamps, output)

    # TODO: test correctness per hand
    # [0,1]
    # {1, 7}
    # {2, 10}
    # [0,2]
    # {1, 8}
    # {2, 11}
    # [0,3]
    # {1, 9}
    # {2, 12}
    # [1,1]
    # {0, 4}
    # [1,2]
    # {0, 5}
    # [1,3]
    # {0, 6}
    # [2,3]
    # {1, 2, 0}

    print_time_max_cliques(time_max_cliques)



def test_avg_alltime():

    filename = "own_files/avg_alltime.txt"
    delta = 0
    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    temporal_graph.mode = 4     # average_alltime
    temporal_graph.del_small_clique = 1

    c = 4.01
    timestamps = -1
    output = False
    time_max_cliques = algorithm(temporal_graph, c, timestamps, output)

    solution_1 = []
    solution_1.append(Solution(0, 3, {1, 2, 0}))
    solution_1.append(Solution(1, 1, {0, 3}))
    solution_1.append(Solution(1, 1, {0, 4}))
    solution_1.append(Solution(1, 1, {0, 5}))
    solution_1.append(Solution(2, 2, {2, 11}))
    solution_1.append(Solution(2, 2, {1, 7}))
    solution_1.append(Solution(2, 2, {1, 8}))
    solution_1.append(Solution(2, 2, {1, 6}))
    solution_1.append(Solution(2, 3, {2, 12}))
    solution_1.append(Solution(2, 3, {2, 10}))
    solution_1.append(Solution(2, 3, {2, 9}))


    assert_temporal_cliques_equal(solution_1, time_max_cliques)

@pytest.mark.parametrize("c,solution", [
    (2.01, [
        Solution(0, 3, {1, 2, 0}),
    ]),
    (12.01, [
        Solution(0, 3, {1, 2, 0}),
        Solution(1, 1, {2, 3}),
        Solution(1, 1, {2, 4}),
        Solution(1, 1, {2, 5}),
        Solution(1, 1, {2, 6}),
        Solution(1, 1, {2, 7}),
        Solution(1, 1, {2, 8}),
    ]),
])
def test_usally_avg(c, solution):
    filename = "own_files/usally_avg.txt"
    delta = 0
    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    temporal_graph.mode = 2     # usally_avg
    temporal_graph.del_small_clique = 1

    timestamps = -1
    output = False
    time_max_cliques = algorithm(temporal_graph, c, timestamps, output)

    assert_temporal_cliques_equal(solution, time_max_cliques)



@pytest.mark.parametrize("c,solution", [
    (2.01, [
        Solution(0, 2, {0, 3}),
        Solution(2, 3, {1, 0}),
        Solution(3, 3, {0, 1, 2}),
    ]),
    (3.01, [
        Solution(0, 2, {0, 3}),
        Solution(0, 2, {1, 4}),
        Solution(0, 2, {1, 5}),
        Solution(0, 3, {0, 1}),
        Solution(2, 3, {0, 1, 2}),
        Solution(3, 3, {2, 8}),
        Solution(3, 3, {2, 9}),
    ]),
    (5.01, [
        Solution(0, 2, {0, 3}),
        Solution(0, 2, {1, 4}),
        Solution(0, 2, {1, 5}),
        Solution(0, 2, {2, 6}),
        Solution(0, 2, {2, 7}),
        Solution(0, 3, {0, 1, 2}),
        Solution(0, 3, {2, 8}),
        Solution(0, 3, {2, 9}),
    ]),
])
def test_max_usally(c, solution):
    filename = "own_files/max_usally.txt"
    delta = 0
    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    temporal_graph.mode = 1     # max_usally
    temporal_graph.del_small_clique = 1

    timestamps = -1
    output = False
    time_max_cliques = algorithm(temporal_graph, c, timestamps, output)

    assert_temporal_cliques_equal(solution, time_max_cliques)



@pytest.mark.parametrize("c,solution", [
    (2.01, [
        Solution(2, 4, {1, 2, 0}),
        Solution(0, 4, {1, 0}),
        Solution(3, 3, {2, 6}),
    ]),
    (3.01, [
        Solution(0, 4, {1, 0}),
        Solution(1, 4, {1, 2, 0}),
        Solution(2, 2, {2, 5}),
        Solution(2, 3, {2, 6}),
    ]),
    (4.01, [
        Solution(0, 4, {1, 2, 0}),
        Solution(1, 1, {2, 4}),
        Solution(1, 2, {2, 5}),
        Solution(1, 3, {2, 6}),
    ]),
    (5.01, [
        Solution(0, 0, {2, 3}),
        Solution(0, 1, {2, 4}),
        Solution(0, 2, {2, 5}),
        Solution(0, 3, {2, 6}),
        Solution(0, 4, {1, 2, 0}),
    ]),
])
def test_alltime_max(c, solution):
    filename = "own_files/alltime_max.txt"
    delta = 0
    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    temporal_graph.mode = 0     # alltime_max
    temporal_graph.del_small_clique = 1

    timestamps = -1
    output = False
    time_max_cliques = algorithm(temporal_graph, c, timestamps, output)

    assert_temporal_cliques_equal(solution, time_max_cliques)


def print_time_max_cliques(time_max_cliques):
    # type: (collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]]) -> None

    # [a, [b, [{clique1}, ..., {clique_n}]]]

    for a in time_max_cliques.keys():
        for b in time_max_cliques[a].keys():
            interval = True
            for clique in time_max_cliques[a][b]:
                if interval:
                    print("[" + str(a) + "," + str(b) + "]")
                    interval = False
                print_clique(clique)
    print("\n")


def print_time_max_bdk_format(time_max_cliques, reverse_name_space, time_space):

    for a in time_max_cliques.keys():
        for b in time_max_cliques[a].keys():
            for clique in time_max_cliques[a][b]:
                print("{", end='')
                for vertex in clique.values():
                    x = vertex
                    print(str(reverse_name_space[x.name]), end='')
                    break
                for vertex in clique.values():
                    if vertex.name != x.name:
                        print("," + str(reverse_name_space[vertex.name]), end='')
                print("} (" + str(time_space[a]) + "," + str(time_space[b]) + ")", end='\n')






def compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, c, output):

    tmpfile = "tmpgraph.txt"
    cliques = vertex_maximal_cliques(run_stages(temporal_graph, static_graph, c))
    static_graph.graph_to_OCaml_file(tmpfile)
    current_path = "../c-isol-1.0/c-isol -max " + str(c) + " < " + tmpfile
    #  cmd = "../c-isol-1.0/c-isol -max " + str(c) + " < " + tmpfile
    cmd = ["../c-isol-1.0/c-isol", "-max", str(c)]
    with open(tmpfile) as f:
        p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=f, encoding="utf-8")
        ocaml_output, stderr = p.communicate()

    sets = ocaml_output.split("\n")  # type: list[str]
    try:
        sets.remove('')
    except ValueError:
        a = 1

    ocaml_cliques = []  # type: list

    for clique in sets:
        cur = clique.split(" ")
        try:
            cur.remove('')
        except ValueError:
            a = 1
        cur_clique = []
        for element in cur:
            cur_clique.append(int(element))

        if len(cur_clique) > temporal_graph.del_small_clique:
            ocaml_cliques.append(sorted(cur_clique))

    our_cliques = []

    for clique in cliques:
        cur_clique = []
        for vertex in clique.values():
            cur_clique.append(int(vertex.name))
        our_cliques.append(sorted(cur_clique))

    our_cliques = sorted(our_cliques)
    ocaml_cliques = sorted(ocaml_cliques)

    # remove non-vertex maximal cliques
    if len(ocaml_cliques) != 0:
        i = 0
        while i < len(ocaml_cliques):
            j = i+1
            while j < len(ocaml_cliques):
                if set(ocaml_cliques[j]) == set(ocaml_cliques[i]):
                    ocaml_cliques.remove(ocaml_cliques[i])
                    i -= 1
                    break
                elif set(ocaml_cliques[i]) >= set(ocaml_cliques[j]):

                    errorlog = open("oCaml_error.log", "w+")
                    errorlog.write("file: " + str(filename) + "\n")
                    errorlog.write("C:    " + str(c) + "\n")
                    errorlog.write(str(ocaml_cliques[i]) + "\n")
                    errorlog.write(str(ocaml_cliques[j]) + "\n")
                    errorlog.write("----------------------------" + "\n")
                    errorlog.close()

                    ocaml_cliques.remove(ocaml_cliques[j])
                    j -= 1
                elif set(ocaml_cliques[j]) >= set(ocaml_cliques[i]):

                    errorlog = open("oCaml_error.log", "w+")
                    errorlog.write("file: " + str(filename) + "\n")
                    errorlog.write("C:    " + str(c) + "\n")
                    errorlog.write(str(ocaml_cliques[i]) + "\n")
                    errorlog.write(str(ocaml_cliques[j]) + "\n")
                    errorlog.write("----------------------------" + "\n")
                    errorlog.close()

                    ocaml_cliques.remove(ocaml_cliques[i])
                    i -= 1
                    break

                j += 1
            i += 1

    ocaml_cliques = sorted(ocaml_cliques)


    if our_cliques == ocaml_cliques:
        return True
    else:
        if output:
            print("path:" + filename)
            print("Timeframe: " + str(static_graph.start_time) + " " + str(static_graph.end_time))
            print("C: " + str(c))
            print("Our Version:")
            print(our_cliques)
            print("OCaml Version:")
            print(ocaml_cliques)

        return False


def ilp_template(filename, a, b, c, mode, delta, outputname):

    start = time.time()
    output = True
    del_small_clique = 0
    period = 20
    # t_min = 0
    t_min = a
    t_max = b
    # t_max = 50

    graph = first(read_temporal_graph(filename, delta, period))

    graph.mode = mode
    # (graph, a, b, c, mode)

    assert compare_ilp(graph, c, t_min, t_max)



def test_ilp_alltime_max():


    filename = "own_files/alltime_max.txt"
    c = 2.01
    mode = 0
    delta = 0

    del_small_clique = 1
    period = 1
    temporal_graph = first(read_temporal_graph(filename, delta, period))

    temporal_graph.mode = mode

    assert compare_ilp(temporal_graph, c)



def tmp_test():

    filename = "highschool/highschool-2011.txt"
    a = 0
    b = 0
    c = 0.1
    period = 20
    delta = 0
    del_small_clique = 0
    mode = 3
    extrafile = "highschool_ilp.lp"
    graph = first(read_temporal_graph(filename, delta, period))

    stat = graph.getStaticGraph(a, b)
    instance_to_ilp(graph, a, b, c, mode, extrafile)



@pytest.mark.parametrize("filename",
    [os.path.join("own_files", filename) for filename in os.listdir("own_files")]
)
def test_compare_with_ocaml(filename):
    delta = 0   # Taking 1 instead of 0 would imply that two timestamps fall together

    c_values = [0, 1, 2, 5, 10, 15, 20, 30, 40, 60, 100]
    # c_values = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    period = 1
    iter_g = read_temporal_graph(filename, delta, period)

    for temporal_graph in iter_g:

        temporal_graph.del_small_clique = 1

        for a in range(temporal_graph.t_min, temporal_graph.t_max + 1):

            static_graph = temporal_graph.getStaticGraph(a, a)
            static_graph.sort_by_degree()

            for c in c_values:
                assert compare_static_graph_with_ocaml(filename, temporal_graph, static_graph, c, True), \
                    "The instance " + str(filename) + " does not match for both algorithms!"


def create_random_graphs():
    nodes = [10, 12, 13, 15, 17, 20]
    edges = [10, 15, 20, 25, 30, 40, 50]
    noises = [0, 1, 2, 4, 6, 8]
    for n in nodes:
        for m in edges:
            for noise in noises:
                if n*(n-1)/2 > m + noise:
                    print("Create a random graph for parameters:")
                    print("n:     " + str(n) + "     m:     " + str(m) + "     noise: " + str(noise))
                    construct_graph(n, m, noise)


def construct_graph(n, m, noise):
    filename = "random_n_" + str(n) + "_m_" + str(m) + "_noise_" + str(noise) + ".txt"
    path = "own_files/" + filename
    static_graph = StaticGraph()
    static_graph.create_random_graph(n, m, noise)
    static_graph.graph_to_Temp_file(path)


def construct_oCaml_instance(filename, a, b, tmpfile):
    delta = 1
    temporal_graph = read_temporal_graph(filename, delta)
    # @fixme: the next line breaks b/c read_temporal_graph returns an iterator over temporal graphs
    static_graph = temporal_graph.getStaticGraph(a, b, "G_" + str(a) + "_" + str(b))
    static_graph.graph_to_OCaml_file(tmpfile)


def non_vertex_maximal_ocaml_instance():

    filename = "own_files/random_n_10_m_25_noise_4.txt"
    a = 0
    b = 0
    c = 5
    tmpfile = "oCaml_error_instance.txt"
    construct_oCaml_instance(filename, a, b, tmpfile)


def speed_test():

    c = 10
    mode = 4
    filename = "co_presence/tij_pres_LH10.dat"
    timestamp = -1
    delta = 0
    outputname = "a.txt"
    start = time.time()
    output = False
    del_small_clique = 1
    period = 20
    temporal_graph_iter = read_temporal_graph(filename, delta, period, mode, del_small_clique)
    num_cand_set = 0
    num_clique_set = 0
    list_time_max_cliques = []
    num_timestamps = 0
    num_iter_timestamps = 0
    num_graph = 0
    for graph in temporal_graph_iter:
        if output and graph.t_max - graph.t_min > 300:
            print("new_graph_" + str(num_graph) + ": " + str(graph.t_min) + " " + str(graph.t_max))
        num_graph += 1
        num_timestamps += graph.t_max - graph.t_min + 1
        time_max_cliques = algorithm(graph, c, timestamp, output)
        num_iter_timestamps += graph.num_timestamps_iter
        list_time_max_cliques.append(time_max_cliques)
        num_cand_set += graph.num_candidate_sets
        num_clique_set += graph.num_clique_before_time_max


def run_ilp_test_on_server():

    t_min = 200
    t_max = 250
    c_values = []
    # c_values.append(0.1) # Correct
    c_values.append(2.01)
    c_values.append(5.01)
    c_values.append(10.1)
    delta = 0
    filename = "highschool/highschool-2011.txt"
    # modes = [0, 1, 2, 3, 4]
    modes = [1, 2, 0, 4]
    # modes = [3, 4]
    # compare for intervall: [0,0]
    # on mode == 3
    for c in c_values:
        for mode in modes:
            ilp_template(filename, t_min, t_max, c, mode, delta, "testfile.txt")


def pre_process_bkd():

    filename = "co_presence/tij_pres_LH10_altered.dat"
    change_period(filename)


@pytest.mark.slow
def test_compare_to_bk():
    path = "co_presence/tij_pres_LH10_altered.dat"
    delta = 0
    period = 1
    mode = 0
    del_small_clique = 1
    t = -1
    change_period(path)
    filename = "tmp.txt"
    outputfile = "log.txt"
    temporal_graph_iter = read_temporal_graph(filename, delta, period, mode, del_small_clique)
    temporal_graph = next(temporal_graph_iter)
    c = temporal_graph.numVertices
    if next(temporal_graph_iter, None) != None:
        print("Warning! There is a gap in the given graph instance. There may be inaccuracity after the first gap.")
        print("If errors occur, change instance such that there is no gap and run it again.")
    list_time_max_cliques = algorithm_clipped(filename, t, c, mode, delta, outputfile, False, period)
    num_time_max = 0
    for time_max_cliques in list_time_max_cliques:
        for a in time_max_cliques.keys():
            for b in time_max_cliques[a].keys():
                for clique in time_max_cliques[a][b]:
                    num_time_max += 1

    print("Numtimmax: " + str(num_time_max))

    a = os.popen("cat " + str(filename) + "  | python BromKerbosch/temporal-kplexes.py").read().split("\n")
    num_lines = len(a)
    curr_line = 2
    solution = []
    while curr_line < num_lines - 3:
        curr = a[curr_line]
        clique = curr.split(" ")[0][1:-1]
        intervall = curr.split(" ")[1][1:-1]
        start = intervall.split(",")[0]
        end = intervall.split(",")[1]
        clique_format = set()  # type: set(int)
        for element in clique.split(","):
            clique_format.add(temporal_graph.name_space[int(element)])
        if len(clique_format) > 1:
            solution.append(Solution(int(start), int(end), clique_format))
        curr_line += 1

    temporal_graph = temporal_graph
    time_max_cliques = list_time_max_cliques[0]
    for entry in solution:
        try:
            assert solution_has_clique(time_max_cliques, entry.clique, entry.a, entry.b)
        except AssertionError:
            "Error on comparison for file " + str(path)
            print("Couldn't find clique in time_max_cliques:")
            print("[" + str(entry.a) + "," + str(entry.b) + "]")
            print(entry.clique)
            output = "{"
            for element in entry.clique:
                output += str(temporal_graph.reverse_name_space[int(element)]) + ","
            output = output[:-1] + "}"
            print(output)
            raise

    assert (len(solution) == num_time_max), "Solution sizes were different:"
    # os.remove(filename)


@pytest.mark.parametrize("c,solution", [
    (7, [
        Solution(0, 0, {1, 2, 0}),
        Solution(0, 0, {3, 2, 0}),
        Solution(0, 0, {1, 2, 4}),
        Solution(0, 0, {4, 2, 3}),
        Solution(0, 0, {1, 2, 5}),
        Solution(0, 0, {5, 2, 3}),
    ]),
])
def use_for_debug():
    filename = "own_files/example_maximiality.txt"
    solution = []
    solution.append(Solution(0, 0, {1, 2, 0}))
    solution.append(Solution(0, 0, {3, 2, 0}))
    solution.append(Solution(0, 0, {1, 2, 4}))
    solution.append(Solution(0, 0, {4, 2, 3}))
    solution.append(Solution(0, 0, {1, 2, 5}))
    solution.append(Solution(0, 0, {5, 2, 3}))
    delta = 0
    period = 1
    mode = 0
    del_small_clique = 1
    t = -1
    outputfile = "log.txt"
    temporal_graph_list = read_temporal_graph(filename, delta, period, mode, del_small_clique)
    c = temporal_graph_list[0].numVertices
    list_time_max_cliques = algorithm_clipped(filename, t, c, mode, delta, outputfile, False, period)
    time_max_cliques = list_time_max_cliques[0]
    print_time_max_cliques(time_max_cliques)
    assert_temporal_cliques_equal(solution, time_max_cliques)


if __name__ == "__main__":
    sys.exit(main())
