from StaticGraph import StaticGraph
from utils import *
import math

# checks whether the average-condition 1 is true on this clique
def avg_condition1(g, vertex, clique, c):
    # type: (StaticGraph, Vertex,  collections.OrderedDict[int, Vertex], float) -> bool
    if len(g.edgeList[vertex]) < (c + 1) * len(clique) - 1:
        return True
    return False

# checks whether the average-condition 2 is true on this clique
def avg_condition2(g, vertex, clique, c):
    # type: (StaticGraph, Vertex, collections.OrderedDict[int, Vertex], float) -> bool

    # Avg-(2) u has fewer than c*|C| outgoing edges -> otherwise delete
    outgoing = 0
    for neighbor in g.edgeList[vertex]:
        if not clique.keys().__contains__(neighbor.id):
            outgoing += 1

    if outgoing < c * len(clique):
        return True
    return False

# checks whether the average-condition 3 is true on this clique
def avg_condition3(g, vertex, clique, c):
    # type: (StaticGraph, Vertex, collections.OrderedDict[int, Vertex], float) -> bool
    if g.number_of_neighbors_in_set(vertex, clique) >= len(clique) - c - 1:
        return True
    return False

# checks whether the average-condition 4 is true on this clique
def avg_condition4(g, u, clique, c):
    # type: (StaticGraph, Vertex, collections.OrderedDict[int, Vertex], float) -> bool

    # Avg-(4) C^u has less than c*(c+1)*|C^u| outgoing edges, where C^u := {x \in C | x<u} U {u}

    # sort vertices by id for all C_U for u [0,|C|]
    vertices = list(clique.values())
    vertices.sort(key=lambda x: x.id, reverse=False)
    c_u = {}
    for vertex in vertices:
        if vertex.id < u.id:
            c_u[vertex.id] = vertex
        else:
            break
    c_u[u.id] = u
    if not g.outgoing_edges_clique(c_u, clique) < c * (c + 1) * len(c_u):
        return False
    return True

# begin trimming average on give clique and exhaustively apply the conditions 'till clique is trimmed or too small
def trim_average(g, clique, v, c_prim, c, del_small_clique):
    # type: (StaticGraph, collections.OrderedDict[int, Vertex], Vertex, float, float, int) -> collections.OrderedDict[int, Vertex]

    nothing = collections.OrderedDict()

    clique_size = len(clique)

    while c_prim >= 0:
        cur_c = c_prim

        remove_set = set()     # type: set[int]

        for vertex in clique.values():
            # Avg-(1) u has less than (c+1)*|C| + 1 neighbors
            if not avg_condition1(g, vertex, clique, c):
                remove_set.add(vertex.id)
                clique_size -= 1
                if del_small_clique >= clique_size:
                    return nothing
                c_prim -= 1

                if c_prim < 0:
                    return nothing

        for value in remove_set:
            clique.pop(value)

        remove_set = set()

        for vertex in clique.values():
            if vertex.id != v.id:
                # Avg-(2) u has fewer than c*|C| outgoing edges -> otherwise delete
                if not avg_condition2(g, vertex, clique, c):
                    remove_set.add(vertex.id)
                    clique_size -= 1
                    if del_small_clique >= clique_size:
                        return nothing
                    c_prim -= 1
                    if c_prim < 0:
                        return nothing

                # Avg-(3) u has at least |C|-c adjacent vertices in C
                elif not avg_condition3(g, vertex, clique, c):
                    remove_set.add(vertex.id)
                    clique_size -= 1
                    if del_small_clique >= clique_size:
                        return nothing
                    c_prim -= 1
                    if c_prim < 0:
                        return nothing

                # Avg-(4) C^u has less than c*(c+1)*|C^u| outgoing edges, where C^u := {x \in C | x<u} U {u}
                elif not avg_condition4(g, vertex, clique, c):
                    remove_set.add(vertex.id)
                    clique_size -= 1
                    if del_small_clique >= clique_size:
                        return nothing
                    c_prim -= 1
                    if c_prim < 0:
                        return nothing

        for value in remove_set:
            clique.pop(value)

        # If no changes have occurred, we add clique
        if cur_c == c_prim:
            return clique
    if c_prim >= 0:
        return clique
    return nothing

def trimming_stage(g, vertex, c, del_small_clique):
    # type: (StaticGraph, Vertex, float, int) -> collections.OrderedDict[int, Vertex]

    degree = len(g.edgeList[vertex])
    if degree > 0:
        clique = g.inclusive_positive_neighborhood(vertex)
        d = degree + 1 - len(clique)
        c_prim = math.ceil(c - 1 - d)

        return trim_average(g, clique, vertex, c_prim, c, del_small_clique)

    return collections.OrderedDict()

